from matplotlib.pyplot import figure
from .linearspline import LinearSpline

class CubicSpline(LinearSpline):
    _grad = None # Gradient
    _curv = None # Curvature
    _pieces = None # Spline Pieces
    @property
    def curv(self):
        if self._curv is None:
            num = len(self.x)
            a = [0.0 for i in range(num)]
            b = [1.0 for i in range(num)]
            c = [0.0 for i in range(num)]
            r = [0.0 for i in range(num)]
            for i in range(1, num-1):
                dxA = self.x[i]-self.x[i-1]
                dxB = self.x[i+1]-self.x[i]
                dyA = self.y[i]-self.y[i-1]
                dyB = self.y[i+1]-self.y[i]
                a[i] = dxA/6
                b[i] = (dxA+dxB)/3
                c[i] = dxB/6
                r[i] = dyB/dxB-dyA/dxA
            gm = [0.0 for i in range(num)]
            bt = b[0]
            self._curv = [0.0 for i in range(num)]
            self._curv[0] = r[0]/bt
            for i in range(1, num):
                gm[i] = c[i-1]/bt
                bt = b[i]-a[i]*gm[i]
                self._curv[i] = (r[i]-a[i]*self._curv[i-1])/bt
            for i in range(num-2, 0, -1):
                self._curv[i] -= gm[i+1]*self._curv[i+1]
        return self._curv
    @property
    def grad(self):
        if self._grad is None:
            self._grad = [piece.interpolate_gradient(0.0) for piece in self.pieces]
            self._grad.append(self.pieces[-1].interpolate_gradient(self.pieces[-1].dx))
        return self._grad
    @property
    def pieces(self):
        if self._pieces is None:
            self._pieces = []
            for i in range(len(self.x)-1):
                xa = self.x[i]
                xb = self.x[i+1]
                ya = self.y[i]
                yb = self.y[i+1]
                d2ya = self.curv[i]
                d2yb = self.curv[i+1]
                self._pieces.append(CubicPiece(xa, xb, ya, yb, d2ya, d2yb))
        return self._pieces
    def single_interpolate_curvature(self, x: float):
        if x > self.xmax or x < self.xmin:
            return ValueError('Lookup value not in range.')
        y = None
        for piece in self.pieces:
            if piece.contains(x):
                y = piece.interpolate_curvature(x=x)
                break
        if y is None:
            return ValueError('Lookup value not found.')
        return y
    def plot_curvature(self, num: int, ax=None, **kwargs):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        x, d2y = [], []
        for piece in self.pieces:
            x = x + piece.x_list(num)
            d2y = d2y + piece.d2y_list(num)
        ax.plot(x, d2y, **kwargs)
        return ax
    def __str__(self):
        from py2md.classes import MDTable
        table = MDTable()
        table.add_column('x', '', data=self.x)
        table.add_column('y', '', data=self.y)
        table.add_column('Gradient', '', data=self.grad)
        table.add_column('Curvature', '', data=self.curv)
        return table.__str__()
    def _repr_markdown_(self):
        return self.__str__()
    def __repr__(self):
        return '<CubicSpline>'

class CubicPiece(object):
    xa = None
    xb = None
    ya = None
    yb = None
    d2ya = None
    d2yb = None
    _dx = None
    _dy = None
    _dydx = None
    def __init__(self, xa: float, xb: float,
                 ya: float, yb: float,
                 d2ya: float, d2yb: float):
        self.xa = xa
        self.xb = xb
        self.ya = ya
        self.yb = yb
        self.d2ya = d2ya
        self.d2yb = d2yb
    @property
    def dx(self):
        if self._dx is None:
            self._dx = self.xb-self.xa
        return self._dx
    @property
    def dy(self):
        if self._dy is None:
            self._dy = self.yb-self.ya
        return self._dy
    @property
    def dydx(self):
        if self._dydx is None:
            self._dydx = self.dy/self.dx
        return self._dydx
    def contains(self, x: float):
        if x >= self.xa and x <= self.xb:
            return True
        else:
            return False
    def interpolate_spline(self, s: float=None, x: float=None):
        if s is None:
            s = x-self.xa
        A = (self.dx-s)/self.dx
        B = s/self.dx
        C = (A**3-A)*self.dx**2/6
        D = (B**3-B)*self.dx**2/6
        return A*self.ya+B*self.yb+C*self.d2ya+D*self.d2yb
    def interpolate_gradient(self, s: float=None, x: float=None):
        if s is None:
            s = x-self.xa
        A = (self.dx-s)/self.dx
        B = s/self.dx
        E = 3*A**2-1
        F = 3*B**2-1
        return self.dydx+(F*self.d2yb-E*self.d2ya)*self.dx/6
    def interpolate_curvature(self, s: float=None, x: float=None):
        if s is None:
            s = x-self.xa
        A = (self.dx-s)/self.dx
        B = s/self.dx
        return A*self.d2ya+B*self.d2yb
    def s_list(self, num: int):
        return [i*self.dx/num for i in range(num+1)]
    def x_list(self, num: int):
        return [self.xa+si for si in self.s_list(num)]
    def y_list(self, num: int):
        return [self.interpolate_spline(s=si) for si in self.s_list(num)]
    def dydx_list(self, num: int):
        return [self.interpolate_gradient(s=si) for si in self.s_list(num)]
    def d2y_list(self, num: int):
        return [self.interpolate_curvature(s=si) for si in self.s_list(num)]
    def __repr__(self):
        return '<CubicPiece>'



class Vector(object):
    """Vector Class"""
    x = None
    y = None
    z = None
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z
    def to_unit(self):
        """Returns the unit vector of this vector"""
        mag = self.return_magnitude()
        if mag == 0:
            return Vector(self.x, self.y, self.z)
        else:
            x = self.x/mag
            y = self.y/mag
            z = self.z/mag
            return Vector(x, y, z)
    def to_point(self):
        """Returns the end point position of this vector"""
        from .point import Point
        return Point(self.x, self.y, self.z)
    def to_vector(self):
        """Returns a copy of this vector"""
        return Vector(self.x, self.y, self.z)
    def return_magnitude(self):
        """Returns the magnitude of this vector"""
        return (self.x**2+self.y**2+self.z**2)**0.5
    def to_xyz(self):
        """Returns the x, y and z values of this vector"""
        return self.x, self.y, self.z
    def __mul__(self, obj):
        from numpy.matlib import matrix
        from pygeom.matrix3d import MatrixVector
        if isinstance(obj, (Vector, MatrixVector)):
            return self.x*obj.x+self.y*obj.y+self.z*obj.z
        elif isinstance(obj, matrix):
            x = self.x*obj
            y = self.y*obj
            z = self.z*obj
            return MatrixVector(x, y, z)
        else:
            x = self.x*obj
            y = self.y*obj
            z = self.z*obj
            return Vector(x, y, z)
    def __rmul__(self, obj):
        from numpy.matlib import matrix
        from pygeom.matrix3d import MatrixVector
        if isinstance(obj, (Vector, MatrixVector)):
            return self.x*obj.x+self.y*obj.y+self.z*obj.z
        elif isinstance(obj, matrix):
            x = obj*self.x
            y = obj*self.y
            z = obj*self.z
            return MatrixVector(x, y, z)
        else:
            x = obj*self.x
            y = obj*self.y
            z = obj*self.z
            return Vector(x, y, z)
    def __truediv__(self, obj):
        if isinstance(obj, (int, float, complex)):
            x = self.x/obj
            y = self.y/obj
            z = self.z/obj
            return Vector(x, y, z)
    def __pow__(self, obj):
        from pygeom.matrix3d import MatrixVector
        if isinstance(obj, Vector):
            x = self.y*obj.z-self.z*obj.y
            y = self.z*obj.x-self.x*obj.z
            z = self.x*obj.y-self.y*obj.x
            return Vector(x, y, z)
        elif isinstance(obj, MatrixVector):
            x = self.y*obj.z-self.z*obj.y
            y = self.z*obj.x-self.x*obj.z
            z = self.x*obj.y-self.y*obj.x
            return MatrixVector(x, y, z)
    def __add__(self, obj):
        from pygeom.matrix3d import MatrixVector
        if isinstance(obj, Vector):
            x = self.x+obj.x
            y = self.y+obj.y
            z = self.z+obj.z
            return Vector(x, y, z)
        elif isinstance(obj, MatrixVector):
            x = self.x+obj.x
            y = self.y+obj.y
            z = self.z+obj.z
            return MatrixVector(x, y, z)
    def __radd__(self, obj):
        if obj == 0 or obj is None:
            return self
        else:
            return self.__add__(obj)
    def __sub__(self, obj):
        from pygeom.matrix3d import MatrixVector
        if isinstance(obj, Vector):
            x = self.x-obj.x
            y = self.y-obj.y
            z = self.z-obj.z
            return Vector(x, y, z)
        elif isinstance(obj, MatrixVector):
            x = self.x-obj.x
            y = self.y-obj.y
            z = self.z-obj.z
            return MatrixVector(x, y, z)
    def __pos__(self):
        return Vector(self.x, self.y, self.z)
    def __neg__(self):
        return Vector(-self.x, -self.y, -self.z)
    def __repr__(self):
        return '<Vector: {:}, {:}, {:}>'.format(self.x, self.y, self.z)
    def __str__(self):
        return '<{:}, {:}, {:}>'.format(self.x, self.y, self.z)
    def __format__(self, format_spec):
        frmstr = '<{:'+format_spec+'}, {:'+format_spec+'}, {:'+format_spec+'}>'
        return frmstr.format(self.x, self.y, self.z)

def vector_from_points(pnta, pntb):
    """Create a Vector from two Points"""
    x = pntb.x-pnta.x
    y = pntb.y-pnta.y
    z = pntb.z-pnta.z
    return Vector(x, y, z)

ihat = Vector(1.0, 0.0, 0.0)
jhat = Vector(0.0, 1.0, 0.0)
khat = Vector(0.0, 0.0, 1.0)
zero_vector = Vector(0.0, 0.0, 0.0)



from pygeom.geom3d import Vector
from sympy import Symbol, sqrt
from sympy import expand, simplify, trigsimp, expand_trig
from sympy import diff, integrate

class SymbolicVector(Vector):
    x = None
    y = None
    z = None
    def __init__(self, x: Symbol, y: Symbol, z: Symbol):
        super(SymbolicVector, self).__init__(x, y, z)
    def return_magnitude(self):
        return sqrt(self.x**2+self.y**2+self.z**2)
    def simplify(self):
        return SymbolicVector(simplify(self.x),
                              simplify(self.y),
                              simplify(self.z))
    def trigsimp(self):
        return SymbolicVector(trigsimp(self.x),
                              trigsimp(self.y),
                              trigsimp(self.z))
    def expand(self):
        return SymbolicVector(expand(self.x),
                              expand(self.y),
                              expand(self.z))
    def expand_trig(self):
        return SymbolicVector(expand_trig(self.x),
                              expand_trig(self.y),
                              expand_trig(self.z))
    def integrate(self, terms):
        return SymbolicVector(integrate(self.x, terms),
                              integrate(self.y, terms),
                              integrate(self.z, terms))
    def diff(self, sym: Symbol):
        return SymbolicVector(diff(self.x, sym),
                              diff(self.y, sym),
                              diff(self.z, sym))
    def subs(self, *args):
        return SymbolicVector(self.x.subs(*args),
                              self.y.subs(*args),
                              self.z.subs(*args))
    def to_unit(self):
        mag = self.return_magnitude()
        return SymbolicVector(self.x/mag, self.y/mag, self.z/mag)
    def __mul__(self, obj):
        if isinstance(obj, SymbolicVector):
            return self.x*obj.x+self.y*obj.y+self.z*obj.z
        else:
            x = obj*self.x
            y = obj*self.y
            z = obj*self.z
            return SymbolicVector(x, y, z)
    def __rmul__(self, obj):
        return self.__mul__(obj)
    def __truediv__(self, obj):
        x = self.x/obj
        y = self.y/obj
        z = self.z/obj
        return SymbolicVector(x, y, z)
    def __pow__(self, obj):
        if isinstance(obj, SymbolicVector):
            x = self.y*obj.z-self.z*obj.y
            y = self.z*obj.x-self.x*obj.z
            z = self.x*obj.y-self.y*obj.x
            return SymbolicVector(x, y, z)
    def __add__(self, obj):
        if isinstance(obj, SymbolicVector):
            x = self.x+obj.x
            y = self.y+obj.y
            z = self.z+obj.z
            return SymbolicVector(x, y, z)
    def __radd__(self, obj):
        if obj == 0 or obj is None:
            return self
        else:
            return self.__add__(obj)
    def __sub__(self, obj):
        if isinstance(obj, SymbolicVector):
            x = self.x-obj.x
            y = self.y-obj.y
            z = self.z-obj.z
            return SymbolicVector(x, y, z)
    def __pos__(self):
        return SymbolicVector(self.x, self.y, self.z)
    def __neg__(self):
        return SymbolicVector(-self.x, -self.y, -self.z)

def symple_vector(label, **kwargs):
    x = Symbol(f'{label:s}.x', **kwargs)
    y = Symbol(f'{label:s}.y', **kwargs)
    z = Symbol(f'{label:s}.z', **kwargs)
    return SymbolicVector(x, y, z)

ihat = SymbolicVector(1, 0, 0)
jhat = SymbolicVector(0, 1, 0)
khat = SymbolicVector(0, 0, 1)



from .vector import Vector, vector_from_points
from math import radians, cos, sin

class Coordinate(object):
    """Coordinate Class"""
    pnt = None
    dirx = None
    diry = None
    dirz = None
    def __init__(self, pnt, vx, vy, vz):
        self.pnt = pnt
        self.dirx = vx.to_unit()
        self.diry = vy.to_unit()
        self.dirz = vz.to_unit()
    def vector_to_global(self, vec):
        """Transforms a vector from this local coordinate system to global"""
        dirx = Vector(self.dirx.x, self.diry.x, self.dirz.x)
        diry = Vector(self.dirx.y, self.diry.y, self.dirz.y)
        dirz = Vector(self.dirx.z, self.diry.z, self.dirz.z)
        x = dirx*vec
        y = diry*vec
        z = dirz*vec
        return Vector(x, y, z)
    def vector_to_local(self, vec):
        """Transforms a vector from global to this local coordinate system"""
        dirx = Vector(self.dirx.x, self.dirx.y, self.dirx.z)
        diry = Vector(self.diry.x, self.diry.y, self.diry.z)
        dirz = Vector(self.dirz.x, self.dirz.y, self.dirz.z)
        x = dirx*vec
        y = diry*vec
        z = dirz*vec
        return Vector(x, y, z)
    def point_to_global(self, pnt):
        """Transforms a point from this local coordinate system to global"""
        vecl = pnt.to_vector()
        vecg = self.vector_to_global(vecl)
        pntg = self.pnt+vecg
        return pntg
    def point_to_local(self, pnt):
        """Transforms a point from global  to this local coordinate system"""
        vecg = vector_from_points(self.pnt, pnt)
        vecl = self.vector_to_local(vecg)
        pntl = vecl.to_point()
        return pntl
    def rotate_about_z(self, angle):
        """Creates a coordinate system that is rotate about x by an angle [deg]"""
        angle = radians(angle)
        cos_ang = cos(angle)
        sin_ang = sin(angle)
        pnt = self.pnt
        dirz = self.dirz
        tx = self.dirx*cos_ang+self.diry*sin_ang
        ty = self.diry*cos_ang-self.dirx*sin_ang
        return Coordinate(pnt, tx, ty, dirz)
    def __repr__(self):
        return '<Coordinate>'

def coordinate_from_points(pnta, pntb, pntc):
    """Create a Coordinate from three Points"""
    pnt = pnta
    dirx = vector_from_points(pnta, pntb).to_unit()
    dirxy = vector_from_points(pnta, pntc).to_unit()
    dirz = dirx**dirxy
    diry = dirz**dirx
    return Coordinate(pnt, dirx, diry, dirz)



from .point import Point
from .vector import Vector, vector_from_points

class Plane(object):
    """Plane Class"""
    pnt = None
    nrm = None
    def __init__(self, pnt, nrm):
        self.pnt = pnt
        self.nrm = nrm.to_unit()
    def return_abcd(self):
        a = self.nrm.x
        b = self.nrm.y
        c = self.nrm.z
        d = -a*self.pnt.x-b*self.pnt.y-c*self.pnt.z
        return a, b, c, d
    def point_z_from_plane(self, pnt):
        vec = vector_from_points(self.pnt, pnt)
        return vec*self.nrm
    def reverse_normal(self):
        self.nrm = -self.nrm
    def __repr__(self):
        return '<Plane>'

def plane_from_multiple_points(pnts):
    """Create a Plane from three Points"""
    n = len(pnts)
    x = [pnt.x for pnt in pnts]
    y = [pnt.y for pnt in pnts]
    z = [pnt.z for pnt in pnts]
    pntc = Point(sum(x)/n, sum(y)/n, sum(z)/n)
    x = [pnt.x-pntc.x for pnt in pnts]
    y = [pnt.y-pntc.y for pnt in pnts]
    z = [pnt.z-pntc.z for pnt in pnts]
    sxx = sum([x[i]**2 for i in range(n)])
    sxy = sum([x[i]*y[i] for i in range(n)])
    sxz = sum([x[i]*z[i] for i in range(n)])
    syy = sum([y[i]**2  for i in range(n)])
    syz = sum([y[i]*z[i] for i in range(n)])
    D = sxx*syy-sxy**2
    a = (syz*sxy-sxz*syy)/D
    b = (sxy*sxz-sxx*syz)/D
    nrm = Vector(a, b, 1.)
    return Plane(pntc, nrm)



from .vector import Vector

class Point(Vector):
    def __init__(self, x, y, z):
        super().__init__(x, y, z)
    def __repr__(self):
        return '<Point: {:}, {:}, {:}>'.format(self.x, self.y, self.z)
    def __str__(self):
        frmstr = '({:}, {:}, {:})'
        return frmstr.format(self.x, self.y, self.z)
    def __format__(self, format_spec: str):
        frmstr = '({:'+format_spec+'}, {:'+format_spec+'}, {:'+format_spec+'})'
        return frmstr.format(self.x, self.y, self.z)

# class Point(object):
#     """Point Class"""
#     x = None
#     y = None
#     z = None
#     def __init__(self, x, y, z):
#         self.x = x
#         self.y = y
#         self.z = z
#     def to_vector(self):
#         """Returns the vector from origin to this point"""
#         from .vector import Vector
#         x = self.x
#         y = self.y
#         z = self.z
#         return Vector(x, y, z)
#     def distance_from_origin(self):
#         """Returns the distance from origin to this point"""
#         vec = self.to_vector()
#         return vec.return_magnitude()
#     def to_xyz(self):
#         """Returns the x, y and z values of this point"""
#         return self.x, self.y, self.z
#     def __eq__(self, obj):
#         if isinstance(obj, Point):
#             return self.x == obj.x and self.y == obj.y and self.z == obj.z
#     def __add__(self, obj):
#         from .vector import Vector
#         if isinstance(obj, Vector):
#             return Point(self.x+obj.x, self.y+obj.y, self.z+obj.z)
#     def __sub__(self, obj):
#         from .vector import Vector
#         if isinstance(obj, Vector):
#             return Point(self.x-obj.x, self.y-obj.y, self.z-obj.z)
#         elif isinstance(obj, Point):
#             return Vector(self.x-obj.x, self.y-obj.y, self.z-obj.z)
#     def __repr__(self):
#         return '<Point: {:}, {:}, {:}>'.format(self.x, self.y, self.z)
#     def __str__(self):
#         frmstr = '({:}, {:}, {:})'
#         return frmstr.format(self.x, self.y, self.z)
#     def __format__(self, format_spec: str):
#         frmstr = '({:'+format_spec+'}, {:'+format_spec+'}, {:'+format_spec+'})'
#         return frmstr.format(self.x, self.y, self.z)

def point_from_lists(x, y, z):
    """Create a list of Point objects"""
    return [Point(x[i], y[i], z[i]) for i in range(len(x))]

def midpoint_of_points(pnts):
    num = len(pnts)
    x = sum(pnt.x for pnt in pnts)/num
    y = sum(pnt.y for pnt in pnts)/num
    z = sum(pnt.z for pnt in pnts)/num
    return Point(x, y, z)



from matplotlib.pyplot import figure

class LinearSpline(object):
    x = None # X Value List
    y = None # Y Value List
    _pieces = None # Spline Pieces
    _xmin = None # Minimum X
    _xmax = None # Maximum X
    def __init__(self, x: list, y: list):
        num = min(len(x), len(y))
        self.x = [x[i] for i in range(num)]
        self.y = [y[i] for i in range(num)]
    @property
    def pieces(self):
        if self._pieces is None:
            self._pieces = []
            for i in range(len(self.x)-1):
                xa = self.x[i]
                xb = self.x[i+1]
                ya = self.y[i]
                yb = self.y[i+1]
                self._pieces.append(LinearPiece(xa, xb, ya, yb))
        return self._pieces
    @property
    def xmin(self):
        if self._xmin is None:
            self._xmin = min(self.x)
        return self._xmin
    @property
    def xmax(self):
        if self._xmax is None:
            self._xmax = max(self.x)
        return self._xmax
    def single_interpolate_spline(self, x: float):
        if x > self.xmax or x < self.xmin:
            return ValueError('Lookup value not in range.')
        y = None
        for piece in self.pieces:
            if piece.contains(x):
                y = piece.interpolate_spline(x=x)
                break
        if y is None:
            return ValueError('Lookup value not found.')
        return y
    def single_interpolate_gradient(self, x: float):
        if x > self.xmax or x < self.xmin:
            return ValueError('Lookup value not in range.')
        dydx = None
        for piece in self.pieces:
            if piece.contains(x):
                dydx = piece.interpolate_gradient(x=x)
                break
        if dydx is None:
            return ValueError('Lookup value not found.')
        return dydx
    def list_interpolation(self, x: list):
        if max(x) > self.xmax or min(x) < self.xmin:
            return ValueError('One or more lookup valus not in range.')
        num = len(self.pieces)
        y = []
        j = 0
        for xi in x:
            found = False
            if not found:
                for i in range(j, num):
                    piece = self.pieces[i]
                    if piece.contains(xi):
                        yi = piece.interpolate_spline(x=xi)
                        j = i
                        found = True
                        break
            if not found:
                for i in range(0, j):
                    piece = self.pieces[i]
                    if piece.contains(xi):
                        yi = piece.interpolate_spline(x=xi)
                        j = i
                        found = True
                        break
            if found:
                y.append(yi)
            else:
                y.append(float('nan'))
        return y
    def plot_spline(self, num: int, ax=None, **kwargs):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        x, y = [], []
        for piece in self.pieces:
            x = x + piece.x_list(num)
            y = y + piece.y_list(num)
        ax.plot(x, y, **kwargs)
        return ax
    def plot_gradient(self, num: int, ax=None, **kwargs):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        x, dydx = [], []
        for piece in self.pieces:
            x = x + piece.x_list(num)
            dydx = dydx + piece.dydx_list(num)
        ax.plot(x, dydx, **kwargs)
        return ax
    def __str__(self):
        from py2md.classes import MDTable
        table = MDTable()
        table.add_column('x', ':', data=self.x)
        table.add_column('y', ':', data=self.y)
        return table.__str__()
    def _repr_markdown_(self):
        return self.__str__()
    def __repr__(self):
        return '<LinearSpline>'

class LinearPiece(object):
    xa = None
    xb = None
    ya = None
    yb = None
    _dx = None
    _dy = None
    _dydx = None
    def __init__(self, xa: float, xb: float, ya: float, yb: float):
        self.xa = xa
        self.xb = xb
        self.ya = ya
        self.yb = yb
    @property
    def dx(self):
        if self._dx is None:
            self._dx = self.xb-self.xa
        return self._dx
    @property
    def dy(self):
        if self._dy is None:
            self._dy = self.yb-self.ya
        return self._dy
    @property
    def dydx(self):
        if self._dydx is None:
            self._dydx = self.dy/self.dx
        return self._dydx
    def contains(self, x: float):
        if x >= self.xa and x <= self.xb:
            return True
        else:
            return False
    def interpolate_spline(self, s: float=None, x: float=None):
        if s is None:
            s = x-self.xa
        return self.ya+self.dydx*s
    def interpolate_gradient(self, s: float=None, x: float=None):
        if s is None:
            s = x-self.xa
        return self.dydx
    def x_list(self):
        return [self.xa, self.xb]
    def y_list(self):
        return [self.ya, self.yb]
    def dydx_list(self):
        return [self.dydx, self.dydx]
    def __repr__(self):
        return '<LinearPiece>'



from .point2d import Point2D
from .vector2d import Vector2D

class InfLine2D(object):
    pnt = None
    uvec = None
    def __init__(self, pnt, uvec):
        self.pnt = pnt
        self.uvec = uvec.to_unit()
    def point_along_vector(self, length):
        x = self.pnt.x+self.uvec.x*length
        y = self.pnt.y+self.uvec.y*length
        return Point2D(x, y)
    def __repr__(self):
        return '<InfLine2D>'

def intersection_of_inflines2D(iln1, iln2, tol=1e-12):
    uv1 = iln1.uvec
    uv2 = iln2.uvec
    xp12 = uv1**uv2
    if abs(xp12) < tol:
        return None
    pt1 = iln1.pnt
    pt2 = iln2.pnt
    x21 = pt2.x-pt1.x
    y21 = pt2.y-pt1.y
    l1 = (uv2.y*x21-uv2.x*y21)/xp12
    xp = pt1.x + l1*uv1.x
    yp = pt1.y + l1*uv1.y
    return Point2D(xp, yp)



from .vector2d import Vector2D

class Point2D(Vector2D):
    def __init__(self, x, y):
        super().__init__(x, y)
    def __repr__(self):
        return '<Point2D: {:}, {:}>'.format(self.x, self.y)
    def __str__(self):
        frmstr = '({:}, {:})'
        return frmstr.format(self.x, self.y)
    def __format__(self, format_spec: str):
        frmstr = '({:'+format_spec+'}, {:'+format_spec+'})'
        return frmstr.format(self.x, self.y)

# class Point2D(object):
#     """Point2D Class"""
#     x = None
#     y = None
#     def __init__(self, x, y):
#         self.x = x
#         self.y = y
#     def to_vector(self):
#         """Returns the vector from origin to this point"""
#         from .vector2d import Vector2D
#         x = self.x
#         y = self.y
#         return Vector2D(x, y)
#     def distance_from_origin(self):
#         """Returns the distance from origin to this point"""
#         vec = self.to_vector()
#         return vec.return_magnitude()
#     def to_xy(self):
#         """Returns the x, y values of this point"""
#         return self.x, self.y
#     def __eq__(self, obj):
#         if isinstance(obj, Point2D):
#             return self.x == obj.x and self.y == obj.y
#     def __add__(self, obj):
#         from .vector2d import Vector2D
#         if isinstance(obj, Vector2D):
#             return Point2D(self.x+obj.x, self.y+obj.y)
#     def __sub__(self, obj):
#         from .vector2d import Vector2D
#         if isinstance(obj, Vector2D):
#             return Point2D(self.x-obj.x, self.y-obj.y)
#         if isinstance(obj, Point2D):
#             return Vector2D(self.x-obj.x, self.y-obj.y)
#     def __repr__(self):
#         return '<Point2D: {:}, {:}>'.format(self.x, self.y)
#     def __str__(self):
#         frmstr = '({:}, {:})'
#         return frmstr.format(self.x, self.y)
#     def __format__(self, format_spec: str):
#         frmstr = '({:'+format_spec+'}, {:'+format_spec+'})'
#         return frmstr.format(self.x, self.y)

def point2d_from_lists(x, y):
    """Create a list of Point2D objects"""
    n = len(x)
    if len(y) == n:
        pnts = [Point2D(x[i], y[i]) for i in range(n)]
    return pnts

def midpoint_of_point2ds(pnts):
    num = len(pnts)
    x = sum(pnt.x for pnt in pnts)/num
    y = sum(pnt.y for pnt in pnts)/num
    return Point2D(x, y)

def point2d_from_complex(z):
    """Create a Point2D from a complex number"""
    x = z.real
    y = z.imag
    return Point2D(x, y)



from .point2d import Point2D
from .vector2d import vector2d_from_points

class Line2D(object):
    """Line2D Class"""
    pnta = None
    pntb = None
    vec = None
    length = None
    def __init__(self, pnta: Point2D, pntb: Point2D):
        self.pnta = pnta
        self.pntb = pntb
        self.update()
    def update(self):
        """Updates this line if the point definition changes"""
        self.vec = vector2d_from_points(self.pnta, self.pntb)
        self.length = self.vec.return_magnitude()
    def centre_point(self):
        """Returns the centre point of this line"""
        x = (self.pnta.x+self.pntb.x)/2
        y = (self.pnta.y+self.pntb.y)/2
        return Point2D(x, y)
    def ratio_point(self, ratio: float):
        """Returns a point a certain ratio along the line"""
        vec = vector2d_from_points(self.pnta, self.pntb)
        return self.pnta+ratio*vec
    def __repr__(self):
        return '<Line2D>'



from .line import Line
from .vector import Vector
from numpy.matlib import zeros
from numpy.linalg import solve
from matplotlib.pyplot import figure
from mpl_toolkits.mplot3d import Axes3D

class CubicSpline(object):
    u"""This class stores a 3D parametric cubic spline."""
    pnts = None
    npnts = None
    clsd = False
    pnls = None
    npnls = None
    d2r = None
    dr = None
    tanA = None
    tanB = None
    R = None
    def __init__(self, pnts, clsd=False, tanA=None, tanB=None):
        u"""This function initialises the object."""
        self.pnts = pnts
        self.clsd = clsd
        self.tanA = tanA
        self.tanB = tanB
        self.update()
    def update(self):
        u"""This function calculates the other parameters of the object."""
        self.npnts = len(self.pnts)
        if self.clsd:
            indas = [i for i in range(self.npnts)]
            indbs = [i+1 for i in range(self.npnts)]
            indbs[-1] = 0
            self.npnls = self.npnts
        else:
            indas = [i for i in range(self.npnts-1)]
            indbs = [i+1 for i in range(self.npnts-1)]
            self.npnls = self.npnts-1
        self.pnls = []
        for i in range(self.npnls):
            inda = indas[i]
            indb = indbs[i]
            self.pnls.append(Line(self.pnts[inda], self.pnts[indb]))
        if self.clsd:
            self.d2r = self.calc_d2r_closed()
            self.dr = self.calc_dr_closed()
        else:
            self.d2r = self.calc_d2r_open(tanA=self.tanA, tanB=self.tanB)
            self.dr = self.calc_dr_open()
#        self.th = self.calc_th()
        self.R = self.calc_R()
    def calc_d2r_open(self, tanA=None, tanB=None):
        u"""This function calculates the curvature of an open ended spline."""
        pnl_dx = [pnl.vec.x/pnl.length for pnl in self.pnls]
        pnl_dy = [pnl.vec.y/pnl.length for pnl in self.pnls]
        pnl_dz = [pnl.vec.z/pnl.length for pnl in self.pnls]
        del_dx = [0.]*self.npnts
        del_dy = [0.]*self.npnts
        del_dz = [0.]*self.npnts
        if tanA != None:
            utA = tanA.to_unit()
            dxA = utA.x
            dyA = utA.y
            dzA = utA.z
            del_dx[0] = pnl_dx[0]-dxA
            del_dy[0] = pnl_dy[0]-dyA
            del_dz[0] = pnl_dz[0]-dzA
        for i in range(1, self.npnts-1):
            del_dx[i] = pnl_dx[i]-pnl_dx[i-1]
            del_dy[i] = pnl_dy[i]-pnl_dy[i-1]
            del_dz[i] = pnl_dz[i]-pnl_dz[i-1]
        if tanB != None:
            utB = tanB.to_unit()
            dxB = utB.x
            dyB = utB.y
            dzB = utB.z
            del_dx[-1] = dxB-pnl_dx[-1]
            del_dy[-1] = dyB-pnl_dy[-1]
            del_dz[-1] = dzB-pnl_dz[-1]
        a = [0.]*self.npnts
        b = [1.]*self.npnts
        c = [0.]*self.npnts
        rx = [0.]*self.npnts
        ry = [0.]*self.npnts
        rz = [0.]*self.npnts
        if tanA != None:
            sB = self.pnls[0].length
            b[0] = sB/3
            c[0] = sB/6
            rx[0] = del_dx[0]
            ry[0] = del_dy[0]
            rz[0] = del_dz[0]
        for i in range(1, self.npnts-1):
            sA = self.pnls[i-1].length
            sB = self.pnls[i].length
            a[i] = sA/6
            b[i] = (sA+sB)/3
            c[i] = sB/6
            rx[i] = del_dx[i]
            ry[i] = del_dy[i]
            rz[i] = del_dz[i]
        if tanB != None:
            sA = self.pnls[-1].length
            a[-1] = sA/6
            b[-1] = sA/3
            rx[-1] = del_dx[-1]
            ry[-1] = del_dy[-1]
            rz[-1] = del_dz[-1]
        Γ = [0.]*self.npnts
        d2x = [0.]*self.npnts
        d2y = [0.]*self.npnts
        d2z = [0.]*self.npnts
        β = b[0]
        d2x[0] = rx[0]/β
        d2y[0] = ry[0]/β
        d2z[0] = rz[0]/β
        for i in range(1, self.npnts):
            Γ[i] = c[i-1]/β
            β = b[i]-a[i]*Γ[i]
            d2x[i] = (rx[i]-a[i]*d2x[i-1])/β
            d2y[i] = (ry[i]-a[i]*d2y[i-1])/β
            d2z[i] = (rz[i]-a[i]*d2z[i-1])/β
        for i in range(self.npnts-2, -1, -1):
            d2x[i] -= Γ[i+1]*d2x[i+1]
            d2y[i] -= Γ[i+1]*d2y[i+1]
            d2z[i] -= Γ[i+1]*d2z[i+1]
        d2r = [Vector(d2x[i], d2y[i], d2z[i]) for i in range(self.npnts)]
        return d2r
    def calc_d2r_closed(self):
        u"""This function calculates the curvature of a closed spline."""
        n = self.npnts
        inda = [i-1 for i in range(n)]
        indb = [i for i in range(n)]
        inda[0] = n-1
        pnl_dx = [pnl.vec.x/pnl.length for pnl in self.pnls]
        pnl_dy = [pnl.vec.y/pnl.length for pnl in self.pnls]
        pnl_dz = [pnl.vec.z/pnl.length for pnl in self.pnls]
        del_dx = [0.]*n
        del_dy = [0.]*n
        del_dz = [0.]*n
        for i in range(n):
            del_dx[i] = pnl_dx[indb[i]]-pnl_dx[inda[i]]
            del_dy[i] = pnl_dy[indb[i]]-pnl_dy[inda[i]]
            del_dz[i] = pnl_dz[indb[i]]-pnl_dz[inda[i]]
        A = zeros((n, n))
        B = zeros((n, 3))
        for i in range(n):
            sA = self.pnls[inda[i]].length
            sB = self.pnls[indb[i]].length
            if i-1 < 0:
                A[i, n-1] = sA/6
            else:
                A[i, i-1] = sA/6
            A[i, i] = (sA+sB)/3
            if i+1 > n-1:
                A[i, 0] = sB/6
            else:
                A[i, i+1] = sB/6
            B[i, 0] = del_dx[i]
            B[i, 1] = del_dy[i]
            B[i, 2] = del_dz[i]
        X = solve(A, B)
        d2x = [X[i, 0] for i in range(n)]
        d2y = [X[i, 1] for i in range(n)]
        d2z = [X[i, 2] for i in range(n)]
        d2r = [Vector(d2x[i], d2y[i], d2z[i]) for i in range(self.npnts)]
        return d2r
    def calc_dr_open(self):
        u"""This function calculates the gradient of an open ended spline."""
        dx = []
        dy = []
        dz = []
        for i in range(self.npnls):
            xA = self.pnts[i].x
            xB = self.pnts[i+1].x
            d2xA = self.d2r[i].x
            d2xB = self.d2r[i+1].x
            yA = self.pnts[i].y
            yB = self.pnts[i+1].y
            d2yA = self.d2r[i].y
            d2yB = self.d2r[i+1].y
            zA = self.pnts[i].z
            zB = self.pnts[i+1].z
            d2zA = self.d2r[i].z
            d2zB = self.d2r[i+1].z
            sP = self.pnls[i].length
            dxA = (xB-xA)/sP-sP/3*d2xA-sP/6*d2xB
            dyA = (yB-yA)/sP-sP/3*d2yA-sP/6*d2yB
            dzA = (zB-zA)/sP-sP/3*d2zA-sP/6*d2zB
            dx.append(dxA)
            dy.append(dyA)
            dz.append(dzA)
        dxB = (xB-xA)/sP+sP/6*d2xA+sP/3*d2xB
        dyB = (yB-yA)/sP+sP/6*d2yA+sP/3*d2yB
        dzB = (zB-zA)/sP+sP/6*d2zA+sP/3*d2zB
        dx.append(dxB)
        dy.append(dyB)
        dz.append(dzB)
        dr = [Vector(dx[i], dy[i], dz[i]) for i in range(self.npnts)]
        return dr
    def calc_dr_closed(self):
        u"""This function calculates the gradient of a closed spline."""
        n = self.npnts
        inda = [i for i in range(n)]
        indb = [i+1 for i in range(n)]
        indb[-1] = 0
        dx = []
        dy = []
        dz = []
        for i in range(self.npnls):
            ia = inda[i]
            ib = indb[i]
            xA = self.pnts[ia].x
            xB = self.pnts[ib].x
            d2xA = self.d2r[ia].x
            d2xB = self.d2r[ib].x
            yA = self.pnts[ia].y
            yB = self.pnts[ib].y
            d2yA = self.d2r[ia].y
            d2yB = self.d2r[ib].y
            zA = self.pnts[ia].z
            zB = self.pnts[ib].z
            d2zA = self.d2r[ia].z
            d2zB = self.d2r[ib].z
            sP = self.pnls[i].length
            dxA = (xB-xA)/sP-sP/3*d2xA-sP/6*d2xB
            dyA = (yB-yA)/sP-sP/3*d2yA-sP/6*d2yB
            dzA = (zB-zA)/sP-sP/3*d2zA-sP/6*d2zB
            dx.append(dxA)
            dy.append(dyA)
            dz.append(dzA)
        dr = [Vector(dx[i], dy[i], dz[i]) for i in range(self.npnts)]
        return dr
    def calc_R(self):
        u"""This function calculates the radius of curvature of the spline."""
        R = []
        for i in range(self.npnts):
            dri = self.dr[i]
            d2ri = self.d2r[i]
            k = (dri**d2ri).return_magnitude()/(dri.return_magnitude())**3
            if k == 0.: 
                R.append(float('inf'))
            else:
                R.append(1/k)
        return R
    def spline_points(self, num=5):
        u"""This function interpolates the spline with a number of points."""
        x = []
        y = []
        z = []
        for i in range(self.npnls):
            ia = i
            xA = self.pnts[ia].x
            d2xA = self.d2r[ia].x
            yA = self.pnts[ia].y
            d2yA = self.d2r[ia].y
            zA = self.pnts[ia].z
            d2zA = self.d2r[ia].z
            ib = i+1
            if ib == self.npnts:
                ib = 0
            xB = self.pnts[ib].x
            d2xB = self.d2r[ib].x
            yB = self.pnts[ib].y
            d2yB = self.d2r[ib].y
            zB = self.pnts[ib].z
            d2zB = self.d2r[ib].z
            sP = self.pnls[i].length
            for j in range(num):
                s = j*sP/num
                A = (sP-s)/sP
                B = s/sP
                C = (A**3-A)*sP**2/6
                D = (B**3-B)*sP**2/6
                x.append(A*xA+B*xB+C*d2xA+D*d2xB)
                y.append(A*yA+B*yB+C*d2yA+D*d2yB)
                z.append(A*zA+B*zB+C*d2zA+D*d2zB)
        if self.clsd:
            x.append(self.pnts[0].x)
            y.append(self.pnts[0].y)
            z.append(self.pnts[0].z)
        else:
            x.append(self.pnts[-1].x)
            y.append(self.pnts[-1].y)
            z.append(self.pnts[-1].z)
        return x, y, z
    def spline_gradient(self, num=5):
        u"""This function interpolates the gradient of the spline."""
        dx = []
        dy = []
        dz = []
        for i in range(self.npnls):
            ia = i
            xA = self.pnts[ia].x
            d2xA = self.d2r[ia].x
            yA = self.pnts[ia].y
            d2yA = self.d2r[ia].y
            zA = self.pnts[ia].z
            d2zA = self.d2r[ia].z
            ib = i+1
            if ib == self.npnts:
                ib = 0
            xB = self.pnts[ib].x
            d2xB = self.d2r[ib].x
            yB = self.pnts[ib].y
            d2yB = self.d2r[ib].y
            zB = self.pnts[ib].z
            d2zB = self.d2r[ib].z
            sP = self.pnls[i].length
            for j in range(num):
                s = j*sP/num
                A = (sP-s)/sP
                B = s/sP
                dx.append((xB-xA)/sP-(3*A**2-1)*sP/6*d2xA+(3*B**2-1)*sP/6*d2xB)
                dy.append((yB-yA)/sP-(3*A**2-1)*sP/6*d2yA+(3*B**2-1)*sP/6*d2yB)
                dz.append((zB-zA)/sP-(3*A**2-1)*sP/6*d2zA+(3*B**2-1)*sP/6*d2zB)
        if self.clsd:
            dx.append(self.dr[0].x)
            dy.append(self.dr[0].y)
            dz.append(self.dr[0].z)
        else:
            dx.append(self.dr[-1].x)
            dy.append(self.dr[-1].y)
            dz.append(self.dr[-1].z)
        return dx, dy, dz
    def spline_curvature(self, num=1):
        u"""This function interpolates the curvature of the spline."""
        d2x = []
        d2y = []
        d2z = []
        for i in range(self.npnls):
            ia = i
            d2xA = self.d2r[ia].x
            d2yA = self.d2r[ia].y
            d2zA = self.d2r[ia].z
            ib = i+1
            if ib == self.npnts:
                ib = 0
            d2xB = self.d2r[ib].x
            d2yB = self.d2r[ib].y
            d2zB = self.d2r[ib].z
            sP = self.pnls[i].length
            for j in range(num):
                s = j*sP/num
                A = (sP-s)/sP
                B = s/sP
                d2x.append(A*d2xA+B*d2xB)
                d2y.append(A*d2yA+B*d2yB)
                d2z.append(A*d2zA+B*d2zB)
        if self.clsd:
            d2x.append(self.d2r[0].x)
            d2y.append(self.d2r[0].y)
            d2z.append(self.d2r[0].z)
        else:
            d2x.append(self.d2r[-1].x)
            d2y.append(self.d2r[-1].y)
            d2z.append(self.d2r[-1].z)
        return d2x, d2y, d2z
    def scatter(self, ax=None, label=False):
        u"""This function plots the points of the spline."""
        if ax == None:
            fig = figure()
            ax = Axes3D(fig)
            ax.grid(True)
        x = []
        y = []
        z = []
        for i in range(self.npnts):
            x.append(self.pnts[i].x)
            y.append(self.pnts[i].y)
            z.append(self.pnts[i].z)
        ax.scatter(x, y, z)
        if label:
            for i in range(self.npnts):
                ax.text(x[i], y[i], z[i], i)
        return ax
    def plot_spline(self, num=5, ax=None, color='blue'):
        u"""This function plots the spline using the interpolated points."""
        if ax == None:
            fig = figure()
            ax = Axes3D(fig)
            ax.grid(True)
        x, y, z = self.spline_points(num)
        ax.plot(x, y, z, color=color)
        return ax
    def arc_length(self, num=1):
        u"""This function calculates the arc length of the spline."""
        s = []
        sc = 0.
        for i in range(self.npnls):
            sP = self.pnls[i].length
            for j in range(num):
                sj = j*sP/num
                s.append(sc+sj)
            sc += sP
        s.append(sc)
        return s
    def plot_gradient(self, ax=None, num=5):
        u"""This function plots the gradient of the spline."""
        if ax == None:
            fig = figure()
            ax = fig.gca()
            ax.grid(True)
            ax.set_title('Gradient')
        s = self.arc_length(num=num)
        dx, dy, dz = self.spline_gradient(num=num)
        ax.plot(s, dx, color='blue')
        ax.plot(s, dy, color='red')
        ax.plot(s, dz, color='green')
        return ax
    def quiver_tangent(self, ax=None, length=1., color='green'):
        u"""This function quiver plots the tangent of the spline."""
        if ax == None:
            fig = figure()
            ax = Axes3D(fig)
            ax.grid(True)
        x = [self.pnts[i].x for i in range(self.npnts)]
        y = [self.pnts[i].y for i in range(self.npnts)]
        z = [self.pnts[i].z for i in range(self.npnts)]
        dx = [self.dr[i].x for i in range(self.npnts)]
        dy = [self.dr[i].y for i in range(self.npnts)]
        dz = [self.dr[i].z for i in range(self.npnts)]
        ax.quiver(x, y, z, dx, dy, dz, length=length, color=color)
        return ax
    def plot_curvature(self, ax=None, num=5):
        u"""This function plots the curvature of the spline."""
        if ax == None:
            fig = figure()
            ax = fig.gca()
            ax.grid(True)
            ax.set_title('Curvature')
        s = self.arc_length(num=num)
        d2x, d2y, d2z = self.spline_curvature(num=num)
        ax.plot(s, d2x, color='blue')
        ax.plot(s, d2y, color='red')
        ax.plot(s, d2z, color='green')
        return ax
    def plot_inverse_radius(self, ax=None, num=5):
        u"""This function plots the inverse radius of curvature of the spline."""
        if ax == None:
            fig = figure()
            ax = fig.gca()
            ax.grid(True)
            ax.set_title('Inverse Radius of Curvature')
        s = self.arc_length(num=num)
        d2x, d2y, d2z = self.spline_curvature(num=num)
        dx, dy, dz = self.spline_gradient(num=num)
        nums = len(s)
        k = []
        for i in range(nums):
            dr = Vector(dx[i], dy[i], dz[i])
            d2r = Vector(d2x[i], d2y[i], d2z[i])
            k.append((dr**d2r).return_magnitude()/(dr.return_magnitude())**3)
        ax.plot(s, k, color='blue')
        return ax
    def quiver_normal(self, ax=None, length=1., color='red'):
        u"""This function quiver plots the normal of the spline."""
        if ax == None:
            fig = figure()
            ax = Axes3D(fig)
            ax.grid(True)
        x = [self.pnts[i].x for i in range(self.npnts)]
        y = [self.pnts[i].y for i in range(self.npnts)]
        z = [self.pnts[i].z for i in range(self.npnts)]
        d2x = [self.d2r[i].x for i in range(self.npnts)]
        d2y = [self.d2r[i].y for i in range(self.npnts)]
        d2z = [self.d2r[i].z for i in range(self.npnts)]
        ax.quiver(x, y, z, d2x, d2y, d2z, length=length, color=color)
        return ax
    def print_gradient(self):
        u"""This function prints the gradient of the spline."""
        outstr = '\nGradient\nID\tdxds\tdyds\tdzds'
        print(outstr)
        frmstr = '{:d}\t{:g}\t{:g}\t{:g}'
        for i in range(self.npnts):
            outstr = frmstr.format(i, round(self.dr[i].x, 6), round(self.dr[i].y, 6), round(self.dr[i].z, 6))
            print(outstr)
    def print_curvature(self):
        u"""This function prints the curvature of the spline."""
        outstr = '\nCurvature\nID\td2xds2\td2yds2\td2zds2\tRadius of Curvature'
        print(outstr)
        frmstr = '{:d}\t{:g}\t{:g}\t{:g}\t{:g}'
        for i in range(self.npnts):
            outstr = frmstr.format(i, round(self.d2r[i].x, 6), round(self.d2r[i].y, 6), round(self.d2r[i].z, 6), round(self.R[i], 6))
            print(outstr)



from pygeom.geom2d import Vector2D, Coordinate2D
from numpy.matlib import matrix, zeros, multiply, divide, arctan2, sqrt, square

class MatrixVector2D(object):
    """MatrixVector2D Class"""
    x = None
    y = None
    def __init__(self, x: matrix, y: matrix):
        self.x = x
        self.y = y
    def to_unit(self):
        """Returns the unit matrix vector of this matrix vector"""
        mag = self.return_magnitude()
        return elementwise_divide(self, mag)
    def return_magnitude(self):
        """Returns the magnitude matrix of this matrix vector"""
        return sqrt(square(self.x) + square(self.y))
    def return_angle(self):
        """Returns the angle matrix of this matrix vector from the x axis"""
        return arctan2(self.y, self.x)
    def __getitem__(self, key):
        x = self.x[key]
        y = self.y[key]
        if isinstance(x, matrix) and isinstance(y, matrix):
            output = MatrixVector2D(x, y)
        else:
            output = Vector2D(x, y)
        return output
    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            self.x[key] = value.x
            self.y[key] = value.y
        else:
            raise IndexError()
    @property
    def shape(self):
        return self.x.shape
    def transpose(self):
        x = self.x.transpose()
        y = self.y.transpose()
        return MatrixVector2D(x, y)
    def sum(self, axis=None, dtype=None, out=None):
        x = self.x.sum(axis=axis, dtype=dtype, out=out)
        y = self.y.sum(axis=axis, dtype=dtype, out=out)
        if isinstance(x, matrix) and isinstance(y, matrix):
            return MatrixVector2D(x, y)
        else:
            return Vector2D(x, y)
    def repeat(self, repeats, axis=None):
        x = self.x.repeat(repeats, axis=axis)
        y = self.y.repeat(repeats, axis=axis)
        return MatrixVector2D(x, y)
    def reshape(self, shape, order='C'):
        x = self.x.reshape(shape, order=order)
        y = self.y.reshape(shape, order=order)
        return MatrixVector2D(x, y)
    def tolist(self):
        lst = []
        for i in range(self.shape[0]):
            lst.append([])
            for j in range(self.shape[1]):
                x = self.x[i, j]
                y = self.y[i, j]
                lst[-1].append(Vector2D(x, y))
        return lst
    def copy(self, order='C'):
        x = self.x.copy(order=order)
        y = self.y.copy(order=order)
        return MatrixVector2D(x, y)
    def to_xy(self):
        """Returns the x and y values of this matrix vector"""
        return self.x, self.y
    def __mul__(self, obj):
        if isinstance(obj, (Vector2D, MatrixVector2D)):
            return self.x*obj.x+self.y*obj.y
        else:
            x = self.x*obj
            y = self.y*obj
            return MatrixVector2D(x, y)
    def __rmul__(self, obj):
        if isinstance(obj, (Vector2D, MatrixVector2D)):
            return obj.x*self.x+obj.y*self.y
        else:
            x = obj*self.x
            y = obj*self.y
            return MatrixVector2D(x, y)
    def __truediv__(self, obj):
        if isinstance(obj, (int, float, complex)):
            x = self.x/obj
            y = self.y/obj
            return MatrixVector2D(x, y)
        else:
            raise TypeError()
    def __pow__(self, obj):
        if isinstance(obj, Vector2D):
            z = self.x*obj.y-self.y*obj.x
            return z
        elif isinstance(obj, MatrixVector2D):
            z = self.x*obj.y-self.y*obj.x
            return z
        else:
            raise TypeError()
    def __add__(self, obj):
        if isinstance(obj, (MatrixVector2D, Vector2D)):
            x = self.x+obj.x
            y = self.y+obj.y
            return MatrixVector2D(x, y)
    def __radd__(self, obj):
        if obj is None:
            return self
        else:
            return self.__add__(obj)
    def __sub__(self, obj):
        if isinstance(obj, (MatrixVector2D, Vector2D)):
            x = self.x-obj.x
            y = self.y-obj.y
            return MatrixVector2D(x, y)
        else:
            raise TypeError()
    def __pos__(self):
        return MatrixVector2D(self.x, self.y)
    def __neg__(self):
        return MatrixVector2D(-self.x, -self.y)
    def __repr__(self):
        return '<MatrixVector2D: {:}, {:}>'.format(self.x, self.y)
    def __str__(self):
        return 'x:\n{:}\ny:\n{:}'.format(self.x, self.y)
    def __format__(self, format_spec):
        frmstr = 'x:\n{:'+format_spec+'}\ny:\n{:'+format_spec+'}'
        return frmstr.format(self.x, self.y)

def zero_matrix_vector(shape: tuple, dtype=float, order='C'):
    x = zeros(shape, dtype=dtype, order=order)
    y = zeros(shape, dtype=dtype, order=order)
    return MatrixVector2D(x, y)

def solve_matrix_vector(a: matrix, b: MatrixVector2D):
    from numpy.linalg import solve
    newb = zeros((b.shape[0], b.shape[1]*2))
    for i in range(b.shape[1]):
        newb[:, 2*i+0] = b[:, i].x
        newb[:, 2*i+1] = b[:, i].y
    newc = solve(a, newb)
    c = zero_matrix_vector(b.shape)
    for i in range(b.shape[1]):
        c[:, i] = MatrixVector2D(newc[:, 2*i+0], newc[:, 2*i+1])
    return c

def elementwise_multiply(a: MatrixVector2D, b: matrix) -> MatrixVector2D:
    if a.shape == b.shape:
        x = multiply(a.x, b)
        y = multiply(a.y, b)
        return MatrixVector2D(x, y)
    else:
        raise ValueError()

def elementwise_divide(a: MatrixVector2D, b: matrix) -> MatrixVector2D:
    if a.shape == b.shape:
        x = divide(a.x, b)
        y = divide(a.y, b)
        return MatrixVector2D(x, y)
    else:
        raise ValueError()

def elementwise_dot_product(a: MatrixVector2D, b: MatrixVector2D) -> matrix:
    if a.shape == b.shape:
        return multiply(a.x, b.x) + multiply(a.y, b.y)
    else:
        raise ValueError()

def elementwise_cross_product(a: MatrixVector2D, b: MatrixVector2D) -> matrix:
    if a.shape == b.shape:
        z = multiply(a.x, b.y)-multiply(a.y, b.x)
        return z
    else:
        raise ValueError()

def vector2d_to_global(self: Coordinate2D, vec: MatrixVector2D) -> MatrixVector2D:
    """Transforms a vector from this local coordinate system to global"""
    dirx = Vector2D(self.dirx.x, self.diry.x)
    diry = Vector2D(self.dirx.y, self.diry.y)
    x = dirx*vec
    y = diry*vec
    return MatrixVector2D(x, y)

def vector2d_to_local(self: Coordinate2D, vec: MatrixVector2D) -> MatrixVector2D:
    """Transforms a vector from global  to this local coordinate system"""
    dirx = self.dirx
    diry = self.diry
    x = dirx*vec
    y = diry*vec
    return MatrixVector2D(x, y)



from .matrixvector2d import MatrixVector2D, zero_matrix_vector, solve_matrix_vector
from .matrixvector2d import elementwise_dot_product, elementwise_cross_product
from .matrixvector2d import elementwise_multiply, elementwise_divide
from .matrixvector2d import vector2d_to_global, vector2d_to_local



from .line2d import Line2D
from .point2d import Point2D
from .vector2d import Vector2D
from numpy.matlib import zeros
from numpy.linalg import solve
from matplotlib.pyplot import figure
from py2md.classes import MDTable

class CubicSpline2D(object):
    u"""This class stores a 2D parametric cubic spline."""
    pnts = None
    npnts = None
    clsd = False
    pnls = None
    npnls = None
    d2r = None
    dr = None
    tanA = None
    tanB = None
    R = None
    def __init__(self, pnts, clsd=False, tanA=None, tanB=None):
        u"""This function initialises the object."""
        self.pnts = pnts
        self.clsd = clsd
        self.tanA = tanA
        self.tanB = tanB
        self.update()
    def update(self):
        u"""This function calculates the other parameters of the object."""
        self.npnts = len(self.pnts)
        if self.clsd:
            indas = [i for i in range(self.npnts)]
            indbs = [i+1 for i in range(self.npnts)]
            indbs[-1] = 0
            self.npnls = self.npnts
        else:
            indas = [i for i in range(self.npnts-1)]
            indbs = [i+1 for i in range(self.npnts-1)]
            self.npnls = self.npnts-1
        self.pnls = []
        for i in range(self.npnls):
            inda = indas[i]
            indb = indbs[i]
            self.pnls.append(SplineLine2D(self.pnts[inda], self.pnts[indb]))
        if self.clsd:
            self.d2r = self.calc_d2r_closed()
            self.dr = self.calc_dr_closed()
        else:
            self.d2r = self.calc_d2r_open(tanA=self.tanA, tanB=self.tanB)
            self.dr = self.calc_dr_open()
        self.R = self.calc_R()
        for i, pnl in enumerate(self.pnls):
            ia, ib = i, i+1
            if ib == self.npnts and self.clsd:
                ib = 0
            pnl.set_d2r(self.d2r[ia], self.d2r[ib])
    def calc_d2r_open(self, tanA=None, tanB=None):
        u"""This function calculates the curvature of an open ended spline."""
        pnl_dx = [pnl.vec.x/pnl.length for pnl in self.pnls]
        pnl_dy = [pnl.vec.y/pnl.length for pnl in self.pnls]
        del_dx = [0.0]*self.npnts
        del_dy = [0.0]*self.npnts
        if tanA != None:
            dxA = tanA.x
            dyA = tanA.y
            del_dx[0] = pnl_dx[0]-dxA
            del_dy[0] = pnl_dy[0]-dyA
        for i in range(1, self.npnts-1):
            del_dx[i] = pnl_dx[i]-pnl_dx[i-1]
            del_dy[i] = pnl_dy[i]-pnl_dy[i-1]
        if tanB != None:
            dxB = tanB.x
            dyB = tanB.y
            del_dx[-1] = dxB-pnl_dx[-1]
            del_dy[-1] = dyB-pnl_dy[-1]
        a = [0.0]*self.npnts
        b = [1.0]*self.npnts
        c = [0.0]*self.npnts
        rx = [0.0]*self.npnts
        ry = [0.0]*self.npnts
        if tanA != None:
            sB = self.pnls[0].length
            b[0] = sB/3
            c[0] = sB/6
            rx[0] = del_dx[0]
            ry[0] = del_dy[0]
        for i in range(1, self.npnts-1):
            sA = self.pnls[i-1].length
            sB = self.pnls[i].length
            a[i] = sA/6
            b[i] = (sA+sB)/3
            c[i] = sB/6
            rx[i] = del_dx[i]
            ry[i] = del_dy[i]
        if tanB != None:
            sA = self.pnls[-1].length
            a[-1] = sA/6
            b[-1] = sA/3
            rx[-1] = del_dx[-1]
            ry[-1] = del_dy[-1]
        Γ = [0.]*self.npnts
        d2x = [0.0]*self.npnts
        d2y = [0.0]*self.npnts
        β = b[0]
        d2x[0] = rx[0]/β
        d2y[0] = ry[0]/β
        for i in range(1, self.npnts):
            Γ[i] = c[i-1]/β
            β = b[i]-a[i]*Γ[i]
            d2x[i] = (rx[i]-a[i]*d2x[i-1])/β
            d2y[i] = (ry[i]-a[i]*d2y[i-1])/β
        for i in range(self.npnts-2, -1, -1):
            d2x[i] -= Γ[i+1]*d2x[i+1]
            d2y[i] -= Γ[i+1]*d2y[i+1]
        d2r = [Vector2D(d2x[i], d2y[i]) for i in range(self.npnts)]
        return d2r
    def calc_d2r_closed(self):
        u"""This function calculates the curvature of a closed spline."""
        n = self.npnts
        inda = [i-1 for i in range(n)]
        indb = [i for i in range(n)]
        inda[0] = n-1
        pnl_dx = [pnl.vec.x/pnl.length for pnl in self.pnls]
        pnl_dy = [pnl.vec.y/pnl.length for pnl in self.pnls]
        del_dx = [0.]*n
        del_dy = [0.]*n
        for i in range(n):
            del_dx[i] = pnl_dx[indb[i]]-pnl_dx[inda[i]]
            del_dy[i] = pnl_dy[indb[i]]-pnl_dy[inda[i]]
        A = zeros((n, n))
        B = zeros((n, 2))
        for i in range(n):
            sA = self.pnls[inda[i]].length
            sB = self.pnls[indb[i]].length
            if i-1 < 0:
                A[i, n-1] = sA/6
            else:
                A[i, i-1] = sA/6
            A[i, i] = (sA+sB)/3
            if i+1 > n-1:
                A[i, 0] = sB/6
            else:
                A[i,i+1] = sB/6
            B[i, 0] = del_dx[i]
            B[i, 1] = del_dy[i]
        X = solve(A, B)
        d2x = [X[i, 0] for i in range(n)]
        d2y = [X[i, 1] for i in range(n)]
        d2r = [Vector2D(d2x[i], d2y[i]) for i in range(self.npnts)]
        return d2r
    def calc_dr_open(self):
        u"""This function calculates the gradient of an open ended spline."""
        dx = []
        dy = []
        for i in range(self.npnls):
            xA = self.pnts[i].x
            xB = self.pnts[i+1].x
            d2xA = self.d2r[i].x
            d2xB = self.d2r[i+1].x
            yA = self.pnts[i].y
            yB = self.pnts[i+1].y
            d2yA = self.d2r[i].y
            d2yB = self.d2r[i+1].y
            sP = self.pnls[i].length
            dxA = (xB-xA)/sP-sP/3*d2xA-sP/6*d2xB
            dyA = (yB-yA)/sP-sP/3*d2yA-sP/6*d2yB
            dx.append(dxA)
            dy.append(dyA)
        dxB = (xB-xA)/sP+sP/6*d2xA+sP/3*d2xB
        dyB = (yB-yA)/sP+sP/6*d2yA+sP/3*d2yB
        dx.append(dxB)
        dy.append(dyB)
        dr = [Vector2D(dx[i], dy[i]) for i in range(self.npnts)]
        return dr
    def calc_dr_closed(self):
        u"""This function calculates the gradient of a closed spline."""
        n = self.npnts
        inda = [i for i in range(n)]
        indb = [i+1 for i in range(n)]
        indb[-1] = 0
        dx = []
        dy = []
        for i in range(self.npnls):
            ia = inda[i]
            ib = indb[i]
            xA = self.pnts[ia].x
            xB = self.pnts[ib].x
            d2xA = self.d2r[ia].x
            d2xB = self.d2r[ib].x
            yA = self.pnts[ia].y
            yB = self.pnts[ib].y
            d2yA = self.d2r[ia].y
            d2yB = self.d2r[ib].y
            sP = self.pnls[i].length
            dxA = (xB-xA)/sP-sP/3*d2xA-sP/6*d2xB
            dyA = (yB-yA)/sP-sP/3*d2yA-sP/6*d2yB
            dx.append(dxA)
            dy.append(dyA)
        dr = [Vector2D(dx[i], dy[i]) for i in range(self.npnts)]
        return dr
    def calc_R(self):
        u"""This function calculates the radius of curvature of the spline."""
        R = []
        for i in range(self.npnts):
            dx = self.dr[i].x
            dy = self.dr[i].y
            d2x = self.d2r[i].x
            d2y = self.d2r[i].y
            k = (dx*d2y-dy*d2x)/(dx**2+dy**2)**1.5
            if k == 0.:
                R.append(float('inf'))
            else:
                R.append(1/k)
        return R
    def spline_points(self, num: int):
        u"""This function interpolates the spline with a number of points."""
        x = []
        y = []
        for pnl in self.pnls:
            for i in range(num):
                s = float(i*pnl.length/num)
                xi, yi = pnl.interpolate_point(s)
                x.append(xi)
                y.append(yi)
        if self.clsd:
            x.append(self.pnts[0].x)
            y.append(self.pnts[0].y)
        else:
            x.append(self.pnts[-1].x)
            y.append(self.pnts[-1].y)
        return x, y
    def spline_gradient(self, num: int):
        u"""This function interpolates the gradient of the spline."""
        dx = []
        dy = []
        for pnl in self.pnls:
            for i in range(num):
                s = float(i*pnl.length/num)
                dxi, dyi = pnl.interpolate_gradient(s)
                dx.append(dxi)
                dy.append(dyi)
        if self.clsd:
            dx.append(self.dr[0].x)
            dy.append(self.dr[0].y)
        else:
            dx.append(self.dr[-1].x)
            dy.append(self.dr[-1].y)
        return dx, dy
    def spline_curvature(self, num: int):
        u"""This function interpolates the curvature of the spline."""
        d2x = []
        d2y = []
        for pnl in self.pnls:
            for i in range(num):
                s = float(i*pnl.length/num)
                d2xi, d2yi = pnl.interpolate_curvature(s)
                d2x.append(d2xi)
                d2y.append(d2yi)
        if self.clsd:
            d2x.append(self.d2r[0].x)
            d2y.append(self.d2r[0].y)
        else:
            d2x.append(self.d2r[-1].x)
            d2y.append(self.d2r[-1].y)
        return d2x, d2y
    def line_intersection(self, line, all_roots=False):
        edct = {}
        xP = line.pnt.x
        yP = line.pnt.y
        dxdl = line.uvec.x
        dydl = line.uvec.y
        for i in range(self.npnls):
            ia = i
            xA = self.pnts[ia].x
            yA = self.pnts[ia].y
            d2xA = self.d2r[ia].x
            d2yA = self.d2r[ia].y
            ib = i+1
            if ib == self.npnts:
                ib = 0
            xB = self.pnts[ib].x
            yB = self.pnts[ib].y
            d2xB = self.d2r[ib].x
            d2yB = self.d2r[ib].y
            sP = self.pnls[i].length
            a = -d2xA*dydl/(6*sP) + d2xB*dydl/(6*sP) + d2yA*dxdl/(6*sP) - d2yB*dxdl/(6*sP)
            b = d2xA*dydl/2 - d2yA*dxdl/2
            c = -d2xA*dydl*sP/3 - d2xB*dydl*sP/6 + d2yA*dxdl*sP/3 + d2yB*dxdl*sP/6 + dxdl*yA/sP - dxdl*yB/sP - dydl*xA/sP + dydl*xB/sP
            d = -dxdl*yA + dxdl*yP + dydl*xA - dydl*xP
            s1, s2, s3 = cubic_roots(a, b, c, d)
            if all_roots:
                edct[i] = [s1/sP, s2/sP, s3/sP]
                continue
            elst = []
            if isinstance(s1, float):
                e1 = s1/sP
                if e1 >= 0.0-sP/1000 and e1 <= sP+sP/1000:
                    elst.append(e1)
            if isinstance(s2, float):
                e2 = s2/sP
                if e2 >= 0.0-sP/1000 and e2 <= sP+sP/1000:
                    elst.append(e2)
            if isinstance(s3, float):
                e3 = s3/sP
                if e3 >= -0.000001 and e3 < 1.000001:
                    elst.append(e3)
            if len(elst) > 0:
                edct[i] = elst
        return edct
    def line_intersection_points(self, line, all_roots=False):
        edct = self.line_intersection(line, all_roots=all_roots)
        print(f'edct = {edct}')
        pnts = []
        for i in edct:
            ia = i
            xA = self.pnts[ia].x
            yA = self.pnts[ia].y
            d2xA = self.d2r[ia].x
            d2yA = self.d2r[ia].y
            ib = i+1
            if ib == self.npnts:
                ib = 0
            xB = self.pnts[ib].x
            yB = self.pnts[ib].y
            d2xB = self.d2r[ib].x
            d2yB = self.d2r[ib].y
            sP = self.pnls[i].length
            elst = edct[i]
            for e in elst:
                s = e*sP
                if isinstance(s, float):
                    if s >= 0.0 and s < sP:
                        A = (sP-s)/sP
                        B = s/sP
                        C = sP**2/6*(A**3-A)
                        D = sP**2/6*(B**3-B)
                        x = A*xA+B*xB+C*d2xA+D*d2xB
                        y = A*yA+B*yB+C*d2yA+D*d2yB
                        pnts.append(Point2D(x, y))
        return pnts
    def scatter(self, ax=None, label=False):
        u"""This function plots the points of the spline."""
        if ax == None:
            fig = figure()
            ax = fig.gca()
            ax.set_aspect('equal')
            ax.grid(True)
        x = []
        y = []
        for i in range(self.npnts):
            x.append(self.pnts[i].x)
            y.append(self.pnts[i].y)
        ax.scatter(x,y)
        if label:
            for i in range(self.npnts):
                ax.text(x[i], y[i], i)
        return ax
    def plot_spline(self, num=1, ax=None, color='blue'):
        u"""This function plots the spline using the interpolated points."""
        if ax == None:
            fig = figure()
            ax = fig.gca()
            ax.set_aspect('equal')
            ax.grid(True)
        x, y = self.spline_points(num)
        ax.plot(x, y, color=color)
        return ax
    def arc_length(self, num=1):
        u"""This function calculates the arc length of the spline."""
        s = []
        sc = 0.0
        for i in range(self.npnls):
            sP = self.pnls[i].length
            for j in range(num):
                sj = j*sP/num
                s.append(sc+sj)
            sc += sP
        s.append(sc)
        return s
    def interpolate_spline_points(self, slst: list):
        u"""This function interpolates the spline points based on arc length."""
        s = self.arc_length()
        i = 0
        ploc = []
        sloc = []
        for si in slst:
            found = False
            if not found:
                for j in range(i, self.npnls):
                    if si >= s[j] and si <= s[j+1]:
                        ploc.append(j)
                        sloc.append(si-s[j])
                        i = j
                        break
            if not found:
                for j in range(0, i):
                    if si >= s[j] and si <= s[j+1]:
                        ploc.append(j)
                        sloc.append(si-s[j])
                        i = j
                        break
        x, y = [], []
        for i, si in enumerate(slst):
            s = sloc[i]
            j = ploc[i]
            pnl = self.pnls[j]
            xi, yi = pnl.interpolate_point(s)
            x.append(xi)
            y.append(yi)
        return x, y
    def interpolate_spline_gradient(self, slst: list):
        u"""This function interpolates the spline gradient based on arc length."""
        s = self.arc_length()
        i = 0
        ploc = []
        sloc = []
        for si in slst:
            found = False
            if not found:
                for j in range(i, self.npnls):
                    if si >= s[j] and si <= s[j+1]:
                        ploc.append(j)
                        sloc.append(si-s[j])
                        i = j
                        break
            if not found:
                for j in range(0, i):
                    if si >= s[j] and si <= s[j+1]:
                        ploc.append(j)
                        sloc.append(si-s[j])
                        i = j
                        break
        dx, dy = [], []
        for i, si in enumerate(slst):
            s = sloc[i]
            j = ploc[i]
            pnl = self.pnls[j]
            dxi, dyi = pnl.interpolate_gradient(s)
            dx.append(dxi)
            dy.append(dyi)
        return dx, dy
    def interpolate_spline_curvature(self, slst: list):
        u"""This function interpolates the spline curvature based on arc length."""
        s = self.arc_length()
        i = 0
        ploc = []
        sloc = []
        for si in slst:
            found = False
            if not found:
                for j in range(i, self.npnls):
                    if si >= s[j] and si <= s[j+1]:
                        ploc.append(j)
                        sloc.append(si-s[j])
                        i = j
                        break
            if not found:
                for j in range(0, i):
                    if si >= s[j] and si <= s[j+1]:
                        ploc.append(j)
                        sloc.append(si-s[j])
                        i = j
                        break
        d2x, d2y = [], []
        for i, si in enumerate(slst):
            s = sloc[i]
            j = ploc[i]
            pnl = self.pnls[j]
            d2xi, d2yi = pnl.interpolate_curvature(s)
            d2x.append(d2xi)
            d2y.append(d2yi)
        return d2x, d2y
    def plot_gradient(self, ax=None, num=5):
        u"""This function plots the gradient of the spline."""
        if ax == None:
            fig = figure()
            ax = fig.gca()
            ax.grid(True)
            ax.set_title('Gradient')
        s = self.arc_length(num=num)
        dx, dy = self.spline_gradient(num=num)
        ax.plot(s, dx, color='blue')
        ax.plot(s, dy, color='red')
        return ax
    def quiver_tangent(self, ax=None, color='green'):
        u"""This function quiver plots the tangent of the spline."""
        if ax == None:
            fig = figure()
            ax = fig.gca()
            ax.set_aspect('equal')
            ax.grid(True)
        x = [self.pnts[i].x for i in range(self.npnts)]
        y = [self.pnts[i].y for i in range(self.npnts)]
        dx = [self.dr[i].x for i in range(self.npnts)]
        dy = [self.dr[i].y for i in range(self.npnts)]
        ax.quiver(x, y, dx, dy, color=color)
        return ax
    def plot_curvature(self, ax=None, num=1):
        u"""This function plots the curvature of the spline."""
        if ax == None:
            fig = figure()
            ax = fig.gca()
            ax.grid(True)
            ax.set_title('Curvature')
        s = self.arc_length(num=num)
        d2x, d2y = self.spline_curvature(num=num)
        ax.plot(s, d2x, color='blue')
        ax.plot(s, d2y, color='red')
        return ax
    def plot_inverse_radius(self, ax=None, num=5):
        u"""This function plots the inverse radius of curvature of the spline."""
        if ax == None:
            fig = figure()
            ax = fig.gca()
            ax.grid(True)
            ax.set_title('Inverse Radius of Curvature')
        s = self.arc_length(num=num)
        d2x, d2y = self.spline_curvature(num=num)
        dx, dy = self.spline_gradient(num=num)
        nums = len(s)
        k = [(dx[i]*d2y[i]-dy[i]*d2x[i])/(dx[i]**2+dy[i]**2)**1.5 for i in range(nums)]
        ax.plot(s, k, color='blue')
        return ax
    def quiver_normal(self, ax=None, color='red'):
        u"""This function quiver plots the normal of the spline."""
        if ax == None:
            fig = figure()
            ax = fig.gca()
            ax.set_aspect('equal')
            ax.grid(True)
        x = [self.pnts[i].x for i in range(self.npnts)]
        y = [self.pnts[i].y for i in range(self.npnts)]
        dx = [-self.dr[i].y for i in range(self.npnts)]
        dy = [self.dr[i].x for i in range(self.npnts)]
        ax.quiver(x, y, dx, dy, color=color)
        return ax
    def print_gradient(self, frmstr='.6f'):
        u"""This function prints the gradient of the spline."""
        from math import cos, sin, atan2, pi
        table = MDTable()
        table.add_column('ID', 'd')
        table.add_column('dxds', frmstr)
        table.add_column('dyds', frmstr)
        table.add_column('Tangent X', frmstr)
        table.add_column('Tangent Y', frmstr)
        for i in range(self.npnts):
            dxds = self.dr[i].x
            dyds = self.dr[i].y
            tanx = cos(atan2(dyds, dxds))
            tany = sin(atan2(dyds, dxds))
            table.add_row([i, dxds, dyds, tanx, tany])
        print(table)
    def print_curvature(self):
        u"""This function prints the curvature of the spline."""
        outstr = '\nCurvature\nID\td2xds2\td2yds2\tRadius'
        print(outstr)
        frmstr = '{:d}\t{:g}\t{:g}\t{:g}'
        for i in range(self.npnts):
            outstr = frmstr.format(i, round(self.d2r[i].x, 6), round(self.d2r[i].x, 6), round(self.R[i], 6))
            print(outstr)

class SplineLine2D(Line2D):
    d2ra = None
    d2rb = None
    def __init__(self, pnta: Point2D, pntb: Point2D):
        super(SplineLine2D, self).__init__(pnta, pntb)
    @property
    def xa(self):
        return self.pnta.x
    @property
    def ya(self):
        return self.pnta.y
    @property
    def xb(self):
        return self.pntb.x
    @property
    def yb(self):
        return self.pntb.y
    def set_d2r(self, d2ra: Vector2D, d2rb: Vector2D):
        self.d2ra = d2ra
        self.d2rb = d2rb
    @property
    def d2xa(self):
        return self.d2ra.x
    @property
    def d2ya(self):
        return self.d2ra.y
    @property
    def d2xb(self):
        return self.d2rb.x
    @property
    def d2yb(self):
        return self.d2rb.y
    def interpolate_point(self, s: float):
        sP = self.length
        A = (sP-s)/sP
        B = s/sP
        C = (A**3-A)*sP**2/6
        D = (B**3-B)*sP**2/6
        x = A*self.xa+B*self.xb+C*self.d2xa+D*self.d2xb
        y = A*self.ya+B*self.yb+C*self.d2ya+D*self.d2yb
        return x, y
    def interpolate_gradient(self, s: float):
        sP = self.length
        A = (sP-s)/sP
        B = s/sP
        E = 3*A**2-1
        F = 3*B**2-1
        dx = (self.xb-self.xa)/sP+(F*self.d2xb-E*self.d2xa)*sP/6
        dy = (self.yb-self.ya)/sP+(F*self.d2yb-E*self.d2ya)*sP/6
        return dx, dy
    def interpolate_curvature(self, s: float):
        sP = self.length
        A = (sP-s)/sP
        B = s/sP
        d2x = A*self.d2xa+B*self.d2xb
        d2y = A*self.d2ya+B*self.d2yb
        return d2x, d2y

def cubic_roots(a: float, b: float, c: float, d: float):
    from math import acos, sqrt, cos, pi, copysign
    tmp = a
    a = b/tmp
    b = c/tmp
    c = d/tmp
    Q = (a**2-3*b)/9
    R = (2*a**3-9*a*b+27*c)/54
    Q3 = Q**3
    R2 = R**2
    if R2 < Q3:
        th = acos(R/sqrt(Q3))
        x1 = -2*sqrt(Q)*cos(th/3)-a/3
        x2 = -2*sqrt(Q)*cos((th+2*pi)/3)-a/3
        x3 = -2*sqrt(Q)*cos((th-2*pi)/3)-a/3
    else:
        A = -copysign((abs(R)+sqrt(R2-Q3))**(1/3), R)
        if A == 0.0:
            B = 0.0
        else:
            B = Q/A
        x1 = (A+B)-a/3
        x2 = complex(-0.5*(A+B)-a/3, sqrt(3)/2*(A-B))
        x3 = complex(x2.real, 0.0-x2.imag)
    return x1, x2, x3



from .point import Point
from .vector import vector_from_points

class Line(object):
    """Line Class"""
    pnta = None
    pntb = None
    vec = None
    length = None
    def __init__(self, pnta, pntb):
        self.pnta = pnta
        self.pntb = pntb
        self.update()
    def update(self):
        """Updates this line if the point definition changes"""
        self.vec = vector_from_points(self.pnta, self.pntb)
        self.length = self.vec.return_magnitude()
    def centre_point(self):
        """Returns the centre point of this line"""
        x = (self.pnta.x+self.pntb.x)/2
        y = (self.pnta.y+self.pntb.y)/2
        z = (self.pnta.z+self.pntb.z)/2
        return Point(x, y, z)
    def ratio_point(self, ratio):
        """Returns a point a certain ratio along the line"""
        vec = vector_from_points(self.pnta, self.pntb)
        return self.pnta+ratio*vec
    def __repr__(self):
        return '<Line>'



from .vector2d import Vector2D, vector2d_from_points
from .point2d import Point2D
from math import cos, sin

class Coordinate2D(object):
    """Coordinate2D Class"""
    pnt = None
    dirx = None
    diry = None
    def __init__(self, pnt: Point2D, dirx: Vector2D, diry: Vector2D):
        self.pnt = pnt
        self.dirx = dirx.to_unit()
        self.diry = diry.to_unit()
    def vector_to_global(self, vec):
        """Transforms a vector from this local coordinate system to global"""
        dirx = Vector2D(self.dirx.x, self.diry.x)
        diry = Vector2D(self.dirx.y, self.diry.y)
        x = dirx*vec
        y = diry*vec
        return Vector2D(x, y)
    def vector_to_local(self, vec: Vector2D):
        """Transforms a vector from global  to this local coordinate system"""
        dirx = self.dirx
        diry = self.diry
        x = dirx*vec
        y = diry*vec
        return Vector2D(x, y)
    def point_to_global(self, pnt: Point2D):
        """Transforms a point from this local coordinate system to global"""
        vecl = pnt.toVector()
        vecg = self.vector_to_global(vecl)
        pntg = self.pnt+vecg
        return pntg
    def point_to_local(self, pnt: Point2D):
        """Transforms a point from global  to this local coordinate system"""
        vecg = vector2d_from_points(self.pnt, pnt)
        vecl = self.vector_to_local(vecg)
        pntl = vecl.to_point()
        return pntl
    def __repr__(self):
        return '<Coordinate2D>'

def coordinate2d_from_points(pnta: Point2D, pntb: Point2D):
    """Create a Vector2D from two Point2Ds"""
    pnt = pnta
    dirx = vector2d_from_points(pnta, pntb)
    diry = Vector2D(-dirx.y, dirx.x)
    return Coordinate2D(pnt, dirx, diry)

def coordinate2d_from_angle(pnt: Point2D, angle: float):
    """Create a Vector2D from a Point2D and an Angle"""
    dirx = Vector2D(cos(angle), sin(angle))
    diry = Vector2D(-dirx.y, dirx.x)
    return Coordinate2D(pnt, dirx, diry)



from .matrixvector import MatrixVector, zero_matrix_vector, solve_matrix_vector
from .matrixvector import elementwise_dot_product, elementwise_cross_product
from .matrixvector import elementwise_multiply, elementwise_divide
from .matrixvector import vector_to_global, vector_to_local



from math import atan2, cos, sin

class Vector2D(object):
    """Vector2D Class"""
    x = None
    y = None
    def __init__(self, x, y):
        self.x = x
        self.y = y
    def to_unit(self):
        """Returns the unit vector of this vector"""
        mag = self.return_magnitude()
        if mag == 0:
            return Vector2D(self.x, self.y)
        else:
            x = self.x/mag
            y = self.y/mag
            return Vector2D(x, y)
    def to_point(self):
        """Returns the end point position of this vector"""
        from .point2d import Point2D
        return Point2D(self.x, self.y)
    def to_vector(self):
        """Returns a copy of this vector"""
        return Vector2D(self.x, self.y)
    def rotate(self, rot):
        """Rotates this vector by an input angle in radians"""
        mag = self.return_magnitude()
        ang = self.return_angle()
        x = mag*cos(ang+rot)
        y = mag*sin(ang+rot)
        return Vector2D(x, y)
    def return_angle(self):
        """Returns the angle of this vector from the x axis"""
        return atan2(self.y, self.x)
    def to_complex(self):
        """Returns the complex number of this vector"""
        cplx = self.x+1j*self.y
        return cplx
    def return_magnitude(self):
        """Returns the magnitude of this vector"""
        return (self.x**2+self.y**2)**0.5
    def to_xy(self):
        """Returns the x, y values of this vector"""
        return self.x, self.y
    def __mul__(self, obj):
        from numpy.matlib import matrix
        from pygeom.matrix2d import MatrixVector2D
        if isinstance(obj, (Vector2D, MatrixVector2D)):
            return self.x*obj.x+self.y*obj.y
        elif isinstance(obj, matrix):
            x = self.x*obj
            y = self.y*obj
            return MatrixVector2D(x, y)
        else:
            x = self.x*obj
            y = self.y*obj
            return Vector2D(x, y)
    def __rmul__(self, obj):
        from numpy.matlib import matrix
        from pygeom.matrix2d import MatrixVector2D
        if isinstance(obj, (Vector2D, MatrixVector2D)):
            return self.x*obj.x+self.y*obj.y
        elif isinstance(obj, matrix):
            x = obj*self.x
            y = obj*self.y
            return MatrixVector2D(x, y)
        else:
            x = obj*self.x
            y = obj*self.y
            return Vector2D(x, y)
    def __truediv__(self, obj):
        x = self.x/obj
        y = self.y/obj
        return Vector2D(x, y)
    def __pow__(self, obj):
        from pygeom.matrix2d import MatrixVector2D
        if isinstance(obj, (Vector2D, MatrixVector2D)):
            return self.x*obj.y-self.y*obj.x
    def __add__(self, obj):
        from pygeom.matrix2d import MatrixVector2D
        if isinstance(obj, Vector2D):
            x = self.x+obj.x
            y = self.y+obj.y
            return Vector2D(x, y)
        elif isinstance(obj, MatrixVector2D):
            x = self.x+obj.x
            y = self.y+obj.y
            return MatrixVector2D(x, y)
    def __radd__(self, obj):
        if obj == 0 or obj is None:
            return self
        else:
            return self.__add__(obj)
    def __sub__(self, obj):
        from pygeom.matrix2d import MatrixVector2D
        if isinstance(obj, Vector2D):
            x = self.x-obj.x
            y = self.y-obj.y
            return Vector2D(x, y)
        elif isinstance(obj, MatrixVector2D):
            x = self.x-obj.x
            y = self.y-obj.y
            return MatrixVector2D(x, y)
    def __pos__(self):
        return Vector2D(self.x, self.y)
    def __neg__(self):
        return Vector2D(-self.x, -self.y)
    def __repr__(self):
        return '<Vector2D: {:}, {:}>'.format(self.x, self.y)
    def __str__(self):
        return '<{:}, {:}>'.format(self.x, self.y)
    def __format__(self, format_spec):
        frmstr = '<{:'+format_spec+'}, {:'+format_spec+'}>'
        return frmstr.format(self.x, self.y)

def vector2d_from_complex(cplx):
    """Create a Vector2D from a complex number"""
    x = cplx.real
    y = cplx.imag
    return Vector2D(x, y)

def vector2d_from_points(pnta, pntb):
    """Create a Vector2D from two Point2Ds"""
    x = pntb.x-pnta.x
    y = pntb.y-pnta.y
    return Vector2D(x, y)

def vector2d_from_magang(mag, ang):
    """Create a Vector2D from magnatude and angle from x direction"""
    x = mag*cos(ang)
    y = mag*sin(ang)
    return Vector2D(x, y)

def vector2d_from_lists(x, y):
    """Create a list of Vector2D objects"""
    n = len(x)
    if len(y) == n:
        vecs = [Vector2D(x[i], y[i]) for i in range(n)]
        return vecs

ihat2d = Vector2D(1.0, 0.0)
jhat2d = Vector2D(0.0, 1.0)
zero_vector2d = Vector2D(0.0, 0.0)
